import delune
import os
from ..helpers.dpath import getdir

def __mount__ (app):
    # replica -------------------------------------------------------
    @app.route ("/<alias>/config", methods = ["GET"])
    @app.permission_required (["index", "replica"])
    def config (was, alias):
        fn = getdir ("config", alias)
        return was.response.file (fn, "application/json")

    @app.route ("/<alias>/locks", methods = ["GET"])
    @app.permission_required ("replica")
    def locks (was, alias):
        return was.API ({"locks": delune.get (alias).si.lock.locks ()})

    @app.route ("/<alias>/locks/<name>", methods = ["POST", "DELETE", "OPTIONS"])
    @app.permission_required ("replica")
    def lock (was, alias, name):
        if was.request.command == "post":
            delune.get (alias).si.lock.lock (name)
            return was.API ("201 Created")
        delune.get (alias).si.lock.unlock (name)
        return was.API ("205 No Content")

    @app.route ("/<alias>/commit", methods = ["POST"])
    @app.permission_required ("index")
    def commit (was, alias):
        delune.get (alias).queue.commit ()
        return was.API ("205 No Content")

    @app.route ("/<alias>/rollback", methods = ["POST"])
    @app.permission_required ("index")
    def rollback (was, alias):
        delune.get (alias).queue.rollback ()
        return was.API ("205 No Content")

    # utilities --------------------------------------------------------------

    @app.route ("/<alias>/stem", methods = ["GET", "POST", "OPTIONS"])
    def stem (was, alias, **args):
        q = args.get ("q")
        if not q:
            returnwas.Fault ("400 Bad Request", 'parameter q required', 40003)
        if isinstance (q, str):
            q = q.split (",")
        l = args.get ("lang", 'en')
        return was.API (dict ([(eq, " ".join (delune.stem (alias, eq, l))) for eq in q]))

    @app.route ("/<alias>/analyze", methods = ["GET", "POST", "OPTIONS"])
    def analyze (was, alias, **args):
        q = args.get ("q")
        if not q:
            return was.Fault ("400 Bad Request", 'parameter q required', 40003)
        l = args.get ("lang", 'en')
        return was.API (delune.analyze (alias, q, l))

    # utilities --------------------------------------------------------------

    @app.route ("/<alias>/devices/<group>/<fn>", methods = ["GET"])
    @app.permission_required ("replica")
    def getfile (was, alias, group, fn):
        s = delune.status (alias)
        if group == "primary":
            path = os.path.join (s ["indexdirs"][0], fn)
        else:
            path = os.path.join (s ["indexdirs"][0], group, fn)
        return was.response.file (path)

    @app.route ("/<alias>/devices/<group>/segments/<fn>", methods = ["GET"])
    @app.permission_required ("replica")
    def getsegfile (was, alias, group, fn):
        s = delune.status (alias)
        seg = fn.split (".") [0]
        if group == "primary":
            if seg not in s ["segmentsizes"]:
                return was.Fault ("404 Not Found", "resource not exist", 40401)
            path = os.path.join (s ["segmentsizes"][seg][0], fn)
        else:
            path = os.path.join (s ["indexdirs"][0], group, fn)
        return was.response.file (path)




import delune
import json

def last_modified (alias):
    return delune.get (alias).si.last_updated

def __mount__ (app):
    @app.route ("/<alias>/documents", methods = ["POST", "DELETE", "OPTIONS"])
    @app.permission_required ("index")
    def documents (was, alias, truncate_confirm = "", q = "", lang = "en", analyze = 1):
        if was.request.method == "DELETE":
            if q:
                delune.get (alias).queue (1, json.dumps ({"query": {'qs': q, 'lang': lang, 'analyze': analyze}}))
                return was.API ("202 Accepted")
            elif truncate_confirm != alias:
                return was.Fault ("400 Bad Request", 'parameter truncate_confirm=(alias name) required', 40003)
            delune.get (alias).queue.truncate ()
            return was.API ("202 Accepted")
        delune.get (alias).queue (0, was.request.body)
        return was.API ("202 Accepted")

    @app.route ("/<alias>/documents/<_id>", methods = ["DELETE", "PUT", "OPTIONS"])
    @app.permission_required ("index")
    def cud (was, alias, _id, nthdoc = 0):
        delune.get (alias).queue (1, json.dumps ({"query": {'qs': "_id:" + _id}}))
        if was.request.method == "PUT":
            delune.get (alias).queue (0, was.request.body)
        return was.API ("202 Accepted")

    # -------------------------------------------------------------

    @app.route ("/<alias>/documents/<_id>", methods = ["GET"])
    def get (was, alias, _id, nthdoc = 0):
        return query (was, alias, "_id:" + _id, nth_content = nth_content)

    @app.route ("/<alias>/documents", methods = ["GET", "PUT", "OPTIONS"])
    def query (was, alias, **args):
        q = args.get ("q")
        if not q:
            return was.Fault ("400 Bad Request", 'parameter q required', 40003)

        o = args.get ("offset", 0)
        f = args.get ("limit", 10)
        s = args.get ("sort", "")
        w = args.get ("snippet", 30)
        r = args.get ("partial", "")
        n = args.get ("nth_content", 0)
        l = args.get ("lang", "en")
        a = args.get ("analyze", 1)
        d = args.get ("data", 1)

        pargs = (o, f, s, w, r, n, l, a, d)
        if type (q) is list:
            # put method need not etag
            return was.API ({ eq: delune.query (alias, eq, *pargs, limit = 1) for eq in q })
        r = delune.query (alias, q, *pargs, limit = 1)
        was.response.set_etag ('{}:{}:{}:{}:{}:{}:{}:{}:{}:{}:{}:{}'.format (
            last_modified (alias), alias, q, *pargs
        ))
        return was.API (r)



from rs4 import unistr
import json

class Document:
    def __init__ (self, _id = None, analyzer = None):
        self._id = _id and str (_id) or None
        self.fields = {}
        self.stored = []
        self.summary = None
        self.analyzer = analyzer

    def __enter__ (self):
        return self

    def __exit__ (self, type, value, tb):
        pass

    def add_field (self, name, value, ftype = "Text", lang = "en", option = [], format = None):
        if self._id and name == "_id":
            raise AssertionError ("_id field should be with document initialization")

        if type (option) is str:
            option = option.split ()

        if ftype == "Fnum":
            try:
                format = [int (each) for each in format.split (".", 1)]
                assert len (format) in (1, 2)
            except:
                raise TypeError ("Invalid Format")

            if len (format) == 2:
                a, b = format
            else:
                a, b = fomat [0], 0

            a += (b + 1)
            ftype += ("%d.%d" % (a, b))

        self.fields [name] = (value, ftype, lang, option)

    def set_contents (self, contents):
        self.stored = contents

    def add_content (self, content):
        # any encoding even allowed python objects
        self.stored.append (content)

    def set_snippet (self, text, lang = "en", option = []):
        if type (option) is str:
            option = option.split ()
        # text SHOUL BE string type, DO CEHCK
        self.summary = (text, lang, option)

    def as_dict (self):
        fields = {}
        if self._id:
            fields ['_id'] = str (self._id)

        for k, v in self.fields.items ():
            fields [k] = {'data': v [0], 'type': v [1], 'lang': v [2], 'option': v [3]}

        return {
            'snippet': self.summary,
            'contents': self.stored,
            'fields': fields
        }

    def as_json (self):
        return json.dumps (self.as_dict ())




import delune
import os
from ..helpers import dpath
import codecs
import json

def __mount__ (app):
    @app.route ("", methods = ["GET"])
    @app.permission_required (["replica", "index"])
    def collections (was, alias = None, side_effect = ""):
        return was.API (collections = list (delune.status ().keys ()))

    @app.route ("/<alias>", methods = ["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"])
    @app.permission_required (["replica", "index"])
    def collections2 (was, alias, side_effect = ""):
        fn = dpath.getdir ("config", alias)
        if was.request.method == "GET":
            if not delune.get (alias):
                return was.Fault ("404 Not Found", "resource %s not exist" % alias, 40401)
            status = delune.status (alias)
            conf = dpath.getdir ("config", alias)
            if not os.path.isfile (conf):
                return was.Fault ("404 Not Found", "resource not exist", 40401)
            with codecs.open (conf, "r", "utf8") as f:
                colopt = json.loads (f.read ())
                status ['colopt'] = {
                    'data': colopt,
                    'mtime': os.path.getmtime (conf),
                    'size': os.path.getsize (conf),
                    'path': conf
                }
            return was.API (status)

        if was.request.method == "DELETE":
            if not os.path.isfile (fn):
                return was.Fault ("404 Not Found", "resource not exist", 40401)
            a, b = os.path.split (fn)
            if side_effect.find ("data") != -1:
                newfn = os.path.join (a, "-" + b)
            else:
                newfn = os.path.join (a, "#" + b)
            if os.path.isfile (newfn):
                os.remove (newfn)
            os.rename (fn, newfn)
            was.setlu (delune.SIG_UPD)
            if side_effect.find ("now") != -1:
                delune.close (alias)
                return was.API ("204 No Content")
            return was.API ("202 Accepted")

        if was.request.method == "POST" and delune.get (alias):
            return was.Fault ("406 Conflict", "resource already exists", 40601)

        elif was.request.method in ("PUT", "PATCH") and not delune.get (alias):
            return was.Fault ("404 Not Found", "resource not exist", 40401)

        if was.request.method == "PATCH":
            with open (fn) as f:
                config = json.load (f)
            data = was.request.JSON
            section = data ["section"]
            for k, v in data ["data"].items ():
                if k not in config [section]:
                    return was.Fault ("400 Bad Request", "{} is not propety of {}".format (k, section), 40001)
                config [section][k] = v
        else:
            config = was.request.JSON

        with open (fn, "w") as f:
            json.dump (config, f)

        was.setlu (delune.SIG_UPD)
        if was.request.method == "POST":
            if side_effect == "now":
                dpath.load_data (alias, app.config.numthreads, was.plock)
                return was.API ("201 Created", **config)
            return was.API ("202 Accepted", **config)
        return was.API (**config)



# 2014. 12. 9 by Hans Roh hansroh@gmail.com

__version__ = "0.4b14"

version_info = tuple (map (lambda x: not x.isdigit () and x or int (x),  __version__.split (".")))

from .analyzers import standardAnalyzer
from .searcher import collection
from . import memory
from . import analyzers
from . import ops
import platform
import os
from .searcher import indexer, document
import threading
import importlib
from .cli import local, restful

PID = None
SIG_UPD = "delune.collections"

standard_analyzer = standardAnalyzer.Analyzer

def get_analyzer (name = "standard"):
	mod = importlib.import_module ("delune.analyzers." + name + "Analyzer")
	return getattr (mod, "Analyzer")

collection = collection.Collection
document = document.Document

# Open Modes
WRITE = "w"
APPEND = "w"
MODIFY = "w"
READ = "r"
CREATE = "c"

osbit, __x = platform.architecture()
if osbit == "64bit":
	LIMIT_SEGMENTSIZE = 10000000000	# 2**39, 10GB but max 500GB is possible
else:
	LIMIT_SEGMENTSIZE = 1900000000 # 2**31, 2GB
LIMIT_FILESIZE = int (LIMIT_SEGMENTSIZE * 0.7)

logger = None

class Task:
	def __init__ (self, logger):
		self.logger = logger
		self.lock = threading.RLock ()
		self._d_ = {}

	def cleanup (self):
		with self.lock:
			reactors = self._d_.items ()
		for alias, reactor in reactors:
			self.logger ("...delune closing %s" % alias, "info")
			reactor.close ()

	def __contains__ (self, alias):
		with self.lock:
			return (alias in self._d_)

	def __getattr__ (self, alias):
		return self.get (alias)

	def __delattr__ (self, alias):
		self.resign (alias)

	def __dir__ (self):
		with self.lock:
			return list (self._d_.keys ())

	def __len__ (self):
		with self.lock:
			return len (self._d_)

	def swap (self, a, b):
		with self.lock:
			t = self._d__ [a]
			self._d__ [a] = b
			self._d__ [b] = t

	def assign (self, alias, obj):
		obj.si.set_ident (alias)
		with self.lock:
			self._d_ [alias] = obj

	def resign (self, alias):
		with self.lock:
			try:
				del self._d_ [alias]
			except KeyError:
				pass

	def get (self, alias):
		with self.lock:
			return self._d_.get (alias)

	def collections (self):
		with self.lock:
			return self._d_.keys ()

	def status (self):
		d = {}
		with self.lock:
			for alias, reactor in list (self._d_.items ()):
				s = reactor.status ()
				if not s:
					close (alias)
					continue
				d [alias] = s
		return d


class NoTask:
	def __getattr__ (self, name):
		raise AssertionError ("delune not configured, configure first")

task = NoTask ()

def configure (numthread, logger_t, io_buf_size = 16384, mem_limit = 128, max_segment_size = 0):
	global logger, task, PID

	if isinstance (task, Task):
		return

	PID = os.getpid ()
	logger = logger_t
	if max_segment_size:
		set_max_segment_size (max_segment_size)

	if not memory.isInitialized ():
		memory.initialize (numthread, io_buf_size, mem_limit, "segment", logger_t)
		analyzers.buildFactory (numthread, logger_t)

	task = Task (logger)
	return task

def shutdown ():
	global task

	if task is None:
		return

	if isinstance (task, Task):
		task.cleanup ()

	if memory.isInitialized ():
		analyzers.close ()
		memory.destroy ()

	task = None

# For Skitai
cleanup = shutdown

def set_max_segment_size (mbytes):
	global LIMIT_SEGMENTSIZE, LIMIT_FILESIZE
	LIMIT_SEGMENTSIZE = int (mbytes * 1000000)
	LIMIT_FILESIZE = int (LIMIT_SEGMENTSIZE * 0.7)

def qualify_analyzer (analyzer):
	if memory.isInitialized ():
		return analyzers.checkIn (analyzer)
	return analyzer

# common jobs ----------------------------------------

def assign (alias, obj):
	global task
	task.assign (alias, obj)

def resign (alias):
	global task
	task.resign (alias)

def close (alias, *args, **karg):
	global task
	reactor = task.get (alias)
	if reactor:
		if not reactor.closed:
			reactor.close (*args, **karg)
		task.resign (alias)

def drop (alias):
	try:
		stu = status (alias)
	except KeyError:
		return
	close (alias)
	for data_dir in stu ['indexdirs']:
		for fn in os.listdir (data_dir):
			os.remove (os.path.join (data_dir, fn))

def get (alias):
	global task
	return task.get (alias)

def status (alias = "", *args, **karg):
	global task
	if alias == "":
		return task.status ()
	return task.get (alias).status (*args, **karg)

def refresh (alias, *args, **karg):
	global task
	return task.get (alias).refresh (*args, **karg)


# searcher only jobs	 -----------------------------------
def query (alias, *args, **karg):
	global task
	return task.get (alias).query (*args, **karg)

def delete (alias, *args, **karg):
	global task
	return task.get (alias).delete (*args, **karg)

def remove (alias, *args, **karg):
	global task
	return task.get (alias).remove (*args, **karg)

def fetch (alias, *args, **karg):
	global task
	return task.get (alias).fetch (*args, **karg)

# classifier only jobs ------------------------------------
def stem (alias, *args, **karg):
	global task
	return task.get (alias).stem (*args, **karg)

def analyze (alias, *args, **karg):
	global task
	return task.get (alias).analyze (*args, **karg)

# remote api ------------------------------------

def mount (addr):
	if not isinstance (addr, str) or addr.startswith ("http://") or addr.startswith ("https://"):
		return restful.Delune (addr)
	else:
		return local.Delune (addr)




import skitai
import delune
import sys, os
import atila
from skitai.corequest.dbi import cluster_manager

def __setup__ (pref):
	skitai.register_states (delune.SIG_UPD)
	assert pref.config.resource_dir
	pref.config.resource_dir = os.path.abspath (pref.config.resource_dir)



# Searchable Field Types
TEXT = "Text"
TERMSET = "Term" # TEXT without position infos, cannot phrase search
STRING = "String"
LIST = "List"
FNUM = "Fnum"
COORD = "Coord4" 
COORD4 = "Coord4" # 4 decimals, 10 M precision
COORD6 = "Coord6" # 6 decimals, 10 CM precision
COORD8 = "Coord8" # 8 decimals, 1 MM precision
BIT8  = "Bit8"
BIT16 = "Bit16"
BIT24 = "Bit24"
BIT32 = "Bit32"
BIT40 = "Bit40"
BIT48 = "Bit48"
BIT56 = "Bit56"
BIT64 = "Bit64"
INT8  = "Int8"
INT16 = "Int16"
INT24 = "Int24"
INT32 = "Int32"
INT40 = "Int40"
INT48 = "Int48"
INT56 = "Int56"
INT64 = "Int64"



import sys
import functools
import delune
import types
from pprint import pprint

def _create ():
    return dl.create (colname, data_dirs or [colname], version, **opts)

def _drop (name):
    assert name == colname
    dl.drop (colname, True)

def _qdelete (q):
    with col:
        col.qdelete (q)
        col.commit ()

def _ddelete (docid):
    with col:
        col.delete (docid)
        col.commit ()

def _search (q, offset = 0, limit = 3):
    with col:
        pprint (col.search (q, int (offset), int (limit)))

def _truncate (name):
    assert name == colname
    with col:
       col.truncate (name)

def _index ():
    raise NotImplementedError

def _delete ():
    raise NotImplementedError

#------------------------------------------------
FUNCS = dict (
    before_search = lambda: None,
    search = _search,
    searched = lambda: None,

    before_create = lambda: None,
    create = _create,
    created = lambda: None,

    before_index = lambda: None,
    index = _index,
    indexed = lambda: None,

    before_delete = lambda: None,
    delete = _delete,
    qdelete = _qdelete,
    ddelete = _ddelete,
    deleted = lambda: None,

    before_truncate = lambda: None,
    truncate = _truncate,
    truncated = lambda: None,

    before_drop = lambda: None,
    drop = _drop,
    dropped = lambda: None,
)

NO_OVERIDES = {'search', 'create', 'qdelete', 'ddelete', 'truncate', 'drop'}

def mapname (name):
    def map (f):
        FUNCS [name] = f
    return map

for k, v in FUNCS.items ():
    if k in NO_OVERIDES:
        continue
    setattr (sys.modules [__name__], k, mapname (k))

def sequencial (a, b, c, *args, **kargs):
    FUNCS [a] ()
    r = FUNCS [b] (*args, **kargs)
    FUNCS [c] ()
    return r


#------------------------------------------------
dl, col, colname = None, None, None
opts, data_dirs, version = {}, None, 1

def setup (endpoint, conf = {}):
    global dl, col, opts, data_dirs, version, colname

    server, colname = endpoint.split ('/cols/')
    dl = delune.mount (server)

    if 'data_dirs' in conf:
        data_dirs = conf.pop ('data_dirs')
    if 'version' in conf:
        version = conf.pop ('version')
    opts = conf

    try:
        col = dl.load (colname)
    except NameError:
        pass
    return col

def handle_argv (endpoint, conf = {}):
    setup (endpoint, conf)
    cmd = len (sys.argv) > 1 and sys.argv [1] or '-'

    if not col and cmd not in ('create', 'help'):
        raise NameError ('collection not created')

    if cmd == "index":
        return sequencial ('before_index', 'index', 'indexed')

    elif cmd == "delete":
        return sequencial ('before_delete', cmd, 'deleted')

    # only command line ------------------------------
    elif cmd == "config":
        for k, v in opts.items ():
            col.setopt (k, **v)
        pprint (col.config)

    elif cmd == "inspect":
        pprint (col.status ())

    elif cmd == "create":
        return sequencial ('before_create', 'create', 'created')

    elif cmd == "truncate":
        assert len (sys.argv) > 2, "collection name required"
        return sequencial ('before_truncate', 'truncate', 'truncated', sys.argv [2])

    elif cmd.endswith ("delete"): # qdelete, ddelete
        assert len (sys.argv) > 2, "q or docId required"
        return sequencial ('before_delete', cmd, 'deleted', sys.argv [2])

    elif cmd == "search":
        assert len (sys.argv) > 2, "q required"
        return sequencial ('before_search', 'search', 'searched', *sys.argv [2:])

    elif cmd == "drop":
        assert len (sys.argv) > 2, "collection name required"
        return sequencial ('before_drop', 'drop', 'dropped', sys.argv [2])

    elif cmd == "help":
        print ("Usage:")
        print ("  script.py COMMAND [OPTION, ...]")
        print ("Commands:")
        print ("  create, config, index, delete, qdelete, ddelete, inspect, truncate, drop, search, help")

    return cmd



import delune
from . import collections, collection, documents
from ..helpers import dpath

def __setup__ (app, mntopt):
    app.mount ("/cols", collections, collection, documents)

def __mount__ (app):
    dpath.RESOURCE_DIR = app.config.resource_dir

    @app.route ("", methods = ["GET"])
    @app.permission_required (["replica", "index"])
    def delune_ (was):
        return was.response.API (
            mounted_dir = "@" + app.config.resource_dir.replace (was.env.get ("HOME", ""), ""),
            n_threads = app.config.numthreads
        )



import delune
import os
import json
import codecs

RESOURCE_DIR = None

def getdir (*d):
    global RESOURCE_DIR
    return os.path.join (RESOURCE_DIR, "delune", *d)

def normpath (path):
    if os.name == "nt":
        return path.replace ("/", "\\")
    return path.replace ("\\", "/")

def load_data (alias, numthreads, plock):
    with codecs.open (getdir ("config", alias), "r", "utf8") as f:
        colopt = json.loads (f.read ())
        colopt ['data_dir'] = [getdir ("collections", normpath(d)) for d in colopt ['data_dir']]

    name = "standard"
    if "name" in colopt ["analyzer"]:
        name = colopt ["analyzer"].get ("name")
        del colopt ["analyzer"]["name"]

    analyzer_class = delune.get_analyzer (name)
    try:
        colopt ["analyzer"].pop ("max_terms")
    except KeyError:
        pass
    analyzer = analyzer_class (8, numthreads, **colopt ["analyzer"])

    si = delune.collection (colopt ["data_dir"], delune.READ, analyzer, plock = plock, version = colopt.get ("version", 1))
    actor = si.get_searcher (**colopt.get ('searcher', {}))
    actor.create_queue ()
    delune.assign (alias, actor)


# 2017. 3. 13 by Hans Roh (hansroh@gmail.com)

from atila import Atila
import delune
from rs4 import pathtool
import os
import json
import codecs
import time
import shutil
import time
from exports import apis, admin
from exports.helpers import dpath

app = Atila (__name__)
app.last_maintern = time.time ()

app.mount ("/admin", admin)
app.mount ("/apis", apis)

@app.before_mount
def before_mount (wasc):
	app.config.numthreads = wasc.numthreads
	app.config.plock = wasc.get_lock (__name__)
	permission_check_handler = wasc.app.config.get ("permission_check_handler")
	if permission_check_handler:
		app.permission_check_handler (permission_check_handler)

	delune.configure (app.config.numthreads, wasc.logger.get ("app"), 16384, 128)
	pathtool.mkdir (dpath.getdir ("config"))

	for alias in os.listdir (dpath.getdir ("config")):
		if alias.startswith ("-"): # remove dropped col
			with app.config.plock:
				with codecs.open (dpath.getdir ("config", alias), "r", "utf8") as f:
					colopt = json.loads (f.read ())
				for d in [dpath.getdir ("collections", dpath.normpath(d)) for d in colopt ['data_dir']]:
					if os.path.isdir (d):
						shutil.rmtree (d)
				os.remove (dpath.getdir ("config", alias))
		elif alias.startswith ("#"): # unused col
			continue
		else:
			dpath.load_data (alias, app.config.numthreads, app.config.plock)
	app.store.set (delune.SIG_UPD, time.time ())

@app.umounted
def umounted (wasc):
	delune.shutdown ()

@app.before_request
def before_request (was):
	if was.request.args.get ('alias') and not (was.request.routed.__name__ == "collections2" and was.request.method == "POST"):
		alias = was.request.args.get ('alias')
		if not delune.get (alias):
			return was.response.Fault ("404 Not Found", 40401, "resource %s not exist" % alias)

@app.maintain (1, threading = False)
def maintain_collections (was, now, count):
	configs = os.listdir (dpath.getdir ("config"))
	for alias in configs:
		if os.path.getmtime (dpath.getdir ("config", alias)) <= app.store.get (delune.SIG_UPD, 0):
			continue
		delune.close (alias)
		dpath.load_data (alias, app.config.numthreads, app.config.plock)
		was.setlu (delune.SIG_UPD)
	if was.getlu (delune.SIG_UPD) <= app.store.get (delune.SIG_UPD):
		return

	was.log ('collection changed, maintern ({}th)...' .format (count))
	for alias in configs:
		if alias [0] in "#-" and delune.get (alias [1:]):
			delune.close (alias [1:])
		elif not delune.get (alias):
			dpath.load_data (alias, app.config.numthreads, app.config.plock)
	app.store.set (delune.SIG_UPD, was.getlu (delune.SIG_UPD))

@app.permission_check_handler
def permission_check_handler (was, perms):
	raddr = was.request.get_remote_addr ()
	if raddr == "127.0.0.1":
		return
	allowed = was.app.config.get ("ADMIN_IPS", {})
	if '*' in allowed:
		return
	if raddr in allowed:
		return
	raise was.Error ("403 Permission Denied")

#----------------------------------------------------------------------------

@app.route ("/status")
@app.permission_required (["index", "replica"])
def status (was):
	return was.status ()





from rs4.webkit import siesta
from . import local
from rs4 import logger
import time
from urllib import parse

class ReponseError (Exception):
    pass


def normpath (point):
    if not point.endswith ('/apis'):
        if not point.endswith ('/'):
            point = point + '/'
        point = parse.urljoin (point, 'apis')
    return point

def make_api (point):
    if not isinstance (point, str):
        return point
    else:
        return siesta.API (normpath (point), reraise_http_error = False)

def assert_or_raise (r, expect):
    if r.status_code != expect:
        raise ReponseError (r)
    return r

class Documents (local.Documents):
    def __init__ (self, name, addr, logger):
        self.addr = addr
        self.logger = logger
        self.api = make_api (self.addr)
        self.name = name
        self.counts = [0, 0]

    def __enter__ (self):
        return self

    def __exit__ (self, type, value, tb):
        pass

    def add (self, doc):
        if doc._id:
            return self.update (doc)
        #self.logger ("#{} add document".format (self.counts [0]), "info", self.name)
        self.counts [0] += 1
        r = self.api.cols (self.name).documents.post (doc.as_dict ())
        return assert_or_raise (r, 202)

    def update (self, doc):
        assert doc._id, "_id required"
        #self.logger ("#{} upsert document ({})".format (self.counts [0], doc._id), "info", self.name)
        self.counts [0] += 1
        r = self.api.cols (self.name).documents (doc._id).put (doc.as_dict ())
        return assert_or_raise (r, 202)

    def delete (self, id):
        #self.logger ("#{} delete document ({})".format (self.counts [1], id), "info", self.name)
        self.counts [1] += 1
        r = self.api.cols (self.name).documents (id).delete ()
        return assert_or_raise (r, 202)

    def qdelete (self, q, lang = "en", analyze = 1):
        r = self.api.cols (self.name).documents.delete (q = q, lang = lang, analyze = analyze)
        return assert_or_raise (r, 202)

    def truncate (self, name):
        r = self.api.cols (self.name).documents.delete (truncate_confirm = name)
        return assert_or_raise (r, 202)

    def search (self, q, offset = 0, limit = 10, **karg):
        r = self.api.cols (self.name).documents.get (q = q, offset = offset, limit = limit, **karg)
        return assert_or_raise (r, 200).data

    def get (self, id, **karg):
        r = self.api.cols (self.name).documents.get (q = "_id:{}".format (id), **karg)
        return assert_or_raise (r, 200).data


class Collection (local.Collection):
    def __init__ (self, name, config, addr, logger = None):
        self.addr = addr
        self.logger = logger
        self.api = make_api (self.addr)
        self.name = name
        self.config = config
        self.documents = Documents (self.name, self.addr, logger)

    def _wait (self, flag = True):
        for i in range (120):
            if self.is_active () is flag:
                return True
            else:
                time.sleep (1)
        return False

    def close (self):
        pass

    def status (self):
        return self.api.cols (self.name).get ().data

    def is_active (self):
        return self.name in self.api.cols.get ().data ["collections"]

    def save (self):
        r = self.api.cols (self.name).put (self.config)
        return assert_or_raise (r, 200)

    def commit (self):
        r = self.api.cols (self.name).commit.post ({})
        return assert_or_raise (r, 205)

    def rollback (self):
        r = self.api.cols (self.name).rollback.post ({})
        return assert_or_raise (r, 205)

    def drop (self, include_date = False):
        if not self.is_active ():
            return
        if include_date:
            r = self.api.cols (self.name).delete (side_effect = 'data')
        else:
            r = self.api.cols (self.name).delete ()
        assert_or_raise (r, 202)
        self._wait (False)
        return r


class Delune (local.Delune):
    def __init__ (self, addr):
        self.addr = addr
        self.logger = logger.screen_logger ()
        self.api = make_api (self.addr)

    def lscol (self):
        r = self.api.cols.get ()
        assert_or_raise (r, 200)
        return r.data ["collections"]

    def load (self, name):
        if name not in self.lscol ():
            raise NameError ("collection not found")
        r = self.api.cols (name).config.get ()
        assert_or_raise (r, 200)
        config = r.data
        return Collection (name, config, self.addr, self.logger)

    def create (self, name, data_dirs, version = 1, **kargs):
        if name in self.lscol ():
            raise NameError ("collection exists")
        config = local.make_config (name, data_dirs, version, **kargs)
        r = self.api.cols (name).post (config)
        assert_or_raise (r, 202)
        col = Collection (name, config, self.addr, self.logger)
        assert col._wait (), "Collection creating timeout"
        return col

    def close (self):
        pass

    def _attr_error (self, *args, **kargs):
        raise AttributeError
    index = getdir = _attr_error
